@@ -40,10 +40,10 @@
 }
 
 extern VALUE rb_mRedcarpet;
+extern VALUE rb_cRenderHTML_TOC;
 VALUE rb_mRender;
 VALUE rb_cRenderBase;
 VALUE rb_cRenderHTML;
-VALUE rb_cRenderHTML_TOC;
 VALUE rb_mSmartyPants;
 
 #define buf2str(t) ((t) ? rb_enc_str_new((const char*)(t)->data, (t)->size, opt->active_enc) : Qnil)
@@ -113,9 +113,10 @@ rndr_tablerow(struct buf *ob, const struct buf *text, void *opaque)
 static void
 rndr_tablecell(struct buf *ob, const struct buf *text, int align, void *opaque)
 {
-	VALUE rb_align;
+	VALUE rb_align, rb_header;
+	VALUE rb_callback, rb_callback_arity;
 
-	switch (align) {
+	switch (align & MKD_TABLE_ALIGNMASK) {
 	case MKD_TABLE_ALIGN_L:
 		rb_align = CSTR2SYM("left");
 		break;
@@ -133,7 +134,25 @@ rndr_tablecell(struct buf *ob, const struct buf *text, int align, void *opaque)
 		break;
 	}
 
-	BLOCK_CALLBACK("table_cell", 2, buf2str(text), rb_align);
+	if (align & MKD_TABLE_HEADER) {
+		rb_header = Qtrue;
+	} else {
+		rb_header = Qfalse;
+	}
+
+	struct redcarpet_renderopt *opt = opaque;
+
+	rb_callback = rb_funcall(opt->self, rb_intern("method"), 1, CSTR2SYM("table_cell"));
+
+	rb_callback_arity = rb_funcall(rb_callback, rb_intern("arity"), 0);
+
+	/* For backward compatibility, let's ensure that the erasure with
+	   only two parameters is still supported. */
+	if (FIX2INT(rb_callback_arity) == 3) {
+		BLOCK_CALLBACK("table_cell", 3, buf2str(text), rb_align, rb_header);
+	} else {
+		BLOCK_CALLBACK("table_cell", 2, buf2str(text), rb_align);
+	}
 }
 
 static void
@@ -390,6 +409,7 @@ static VALUE rb_redcarpet_rbase_alloc(VALUE klass)
 static void rb_redcarpet__overload(VALUE self, VALUE base_class)
 {
 	struct rb_redcarpet_rndr *rndr;
+	VALUE options_ivar;
 
 	Data_Get_Struct(self, struct rb_redcarpet_rndr, rndr);
 	rndr->options.self = self;
@@ -411,7 +431,8 @@ static void rb_redcarpet__overload(VALUE self, VALUE base_class)
 		}
 	}
 
-	if (rb_iv_get(self, "@options") == Qnil)
+	options_ivar = rb_attr_get(self, rb_intern("@options"));
+	if (options_ivar == Qundef || options_ivar == Qnil)
 		rb_iv_set(self, "@options", rb_hash_new());
 }
 
@@ -511,10 +532,22 @@ static VALUE rb_redcarpet_htmltoc_init(int argc, VALUE *argv, VALUE self)
 	sdhtml_toc_renderer(&rndr->callbacks, (struct html_renderopt *)&rndr->options.html, render_flags);
 	rb_redcarpet__overload(self, rb_cRenderHTML_TOC);
 
-	if (!(NIL_P(nesting_level)))
-		rndr->options.html.toc_data.nesting_level = NUM2INT(nesting_level);
-	else
-		rndr->options.html.toc_data.nesting_level = 6;
+	/* Check whether we are dealing with a Range object by
+	   checking whether the object responds to min and max */
+	if (rb_respond_to(nesting_level, rb_intern("min")) &&
+	    rb_respond_to(nesting_level, rb_intern("max"))) {
+		int min = NUM2INT(rb_funcall(nesting_level, rb_intern("min"), 0));
+		int max = NUM2INT(rb_funcall(nesting_level, rb_intern("max"), 0));
+
+		rndr->options.html.toc_data.nesting_bounds[0] = min;
+		rndr->options.html.toc_data.nesting_bounds[1] = max;
+	} else if (FIXNUM_P(nesting_level)) {
+		rndr->options.html.toc_data.nesting_bounds[0] = 1;
+		rndr->options.html.toc_data.nesting_bounds[1] = NUM2INT(nesting_level);
+	} else {
+		rndr->options.html.toc_data.nesting_bounds[0] = 1;
+		rndr->options.html.toc_data.nesting_bounds[1] = 6;
+	}
 
 	return Qnil;
 }
